#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "graph.h"
#include "fastme.h"
#include "distance.h"

/*code is heavily borrowed from BIONJ*/


double **initDoubleMatrix(int d);
void freeMatrix(double **D, int size);

tree *Finish(double **delta, int N, int *Emptied, node **nodeArray);
tree *detrifurcate(tree *T);
node *copyNode(node *v);

double Agglomerative_criterion(int i, int j, double **delta, int r);

double Reduction4(int a, double la, int b, double lb, int i,
		  double **delta);
double Branch_length(int a, int b, double **delta, int r);


void copyMatrixEntries(double **X, double **Y, int d)
{
  int i,j;
  for(i=0;i<d;i++)
    for(j=0;j<d;j++)
      Y[i][j] = X[i][j];  
}

void Compute_sums_Sx(double **delta, int n, int *Emptied);
void Best_pair(double **delta, int r, int *a, int *b, int n, int *Emptied);
int *initZeroArray(int l)
{
  int *x;
  int i;
  x = (int *)calloc(l,sizeof(int));
  for(i=0;i<l;i++)
    x[i] = 0;
  return(x);
}

int *initOneArray(int l)
{
  int *x;
  int i;
  x = (int *)calloc(l,sizeof(int));
  for(i=0;i<l;i++)
    x[i] = 1;
  return(x);
}



node **makeArrayFromSet(set *X, int n)
{
  set *looper;
  node **nodeList;
  node *newNode;
  nodeList = (node **)calloc(n,sizeof(node *));
  for(looper=X;NULL!=looper;looper=looper->secondNode)
    {
      newNode = copyNode(looper->firstNode);
      nodeList[looper->firstNode->index2]=newNode;
      newNode->index=looper->firstNode->index2;
    }
  return(nodeList);
}

int support(int *v, int n);

tree *nj(double **D, set *species, int n)
{
  int r,N;
  int *Emptied;
  int a, b;
  double **delta;
  int i;
  double la, lb;
  node *newNode, *nleft, *nright;
  node **nodeArray;
  edge *left,*right;
  int newIndex;
  char edgeLabel[EDGE_LABEL_LENGTH];
  char nodeLabel[NODE_LABEL_LENGTH];
  tree *njTree;
  
  N=2*n-2;
  delta = initDoubleMatrix(N);
  nodeArray = makeArrayFromSet(species,N);
  copyMatrixEntries(D,delta,n); /*copies D into delta*/
  newIndex=n;
  Emptied = initZeroArray(N);
  for(a=n;a<N;a++)
    Emptied[a]=1;
  r = n; 
  
  while (r > 3)                             /* until r=3                 */
    {
      Compute_sums_Sx(delta, N, Emptied);             /* compute the sum Sx       */
      Best_pair(delta, r, &a, &b, N, Emptied);          /* find the best pair by    */
      /* minimizing (1)           */
      /*note: a<b after this step*/
      la=Branch_length(a, b, delta, r);    /* compute branch-lengths   */
      lb=Branch_length(b, a, delta, r);    /* using formula (2)        */
      

      /*tree collapsing*/
      nleft = nodeArray[a];
      nright = nodeArray[b];
      sprintf(nodeLabel,"");
      newNode=makeNewNode(nodeLabel,newIndex);
      newNode->index2=newIndex;
      sprintf(edgeLabel,"E%d",2*(newIndex-n)+1);
      left = makeEdge(edgeLabel,newNode,nleft,la);
      sprintf(edgeLabel,"E%d",2*(newIndex-n)+2);
      right = makeEdge(edgeLabel,newNode,nright,lb);
      newNode->leftEdge = left;
      newNode->rightEdge = right;
      nleft->parentEdge = left;
      nright->parentEdge = right;
      if (0 != Emptied[a])
	{
	  fprintf(stderr,"Error a: NJ using wrong node %d. r is %d\n",a,r);
	  exit(EXIT_FAILURE);
	}
      if (0 != Emptied[b])
	{
	  fprintf(stderr,"Error b: NJ using wrong node %d. r is %d\n",b,r);
	  exit(EXIT_FAILURE);
	}
      Emptied[a]=Emptied[b]=1;
      Emptied[newIndex]=0;
      /*printf("r is %d support of Emptied is %d.\n",r,support(Emptied,N));*/
      nodeArray[newIndex]=newNode;
      
      /*assign distances for newNode*/
      for(i=0; i < N; i++)
	{
	  if(0==Emptied[i])
	    delta[i][newIndex]=delta[newIndex][i]
	      =Reduction4(a, la, b, lb, i, delta);
	  /*use Reduction4 (from BIONJ) to fill in entries for newNode*/
	}
      
      newIndex++;
      r--;                                /* decrease r                */
   }
  
  
  njTree = Finish(delta, N, Emptied, nodeArray);       /* compute the branch-lengths*/
  freeMatrix(delta,N);
  free(Emptied);
  free(nodeArray);
  njTree = detrifurcate(njTree);
  return(njTree);

}
  
/*;;;;;;;;;;;;;;;;;;;;;;;Compute_sums_Sx;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
; Description : This function computes the sums Sx and store them in the    ;
;               diagonal the delta matrix.                                  ;
;                                                                           ;
; input       :                                                             ;
;     	         double **delta : the delta matrix.                      ;
;     	         int n          : the number of taxa                    ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

void Compute_sums_Sx(double **delta, int n, int *Emptied)
{
  double sum;
  int i;
  int j;
  
  for(i= 0; i < n ; i++)
    {
      if(0==Emptied[i])
	{
	  sum=0;
	  for(j=0; j < n; j++)
	    {
	      if(i != j && (0==Emptied[j]))           /* compute the sum Si */
		sum=sum + delta[i][j];
	    }
	}
      delta[i][i]=sum;                           /* store the sum Si in */
    }                                               /* delta diagonal    */
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;Best_pair;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;  Description : This function finds the best pair to be agglomerated by    ;
;                minimizing the agglomerative criterion (1).                ;
;                                                                           ;
;  input       :                                                            ;
;                double **delta : the delta matrix                           ;
;                int r          : number of subtrees                        ;
;                int *a         : contain the first taxon of the pair       ;
;                int *b         : contain the second taxon of the pair      ;
;                int n          : number of taxa                            ;
;                                                                           ;
;  return value:                                                            ;
;                int *a         : the first taxon of the pair               ;
;                int *b         : the second taxon of the pair              ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/


void Best_pair(double **delta, int r, int *a, int *b, int n, int *Emptied)
{
  double Qxy;                         /* value of the criterion calculated*/
  int i,j;                           /* the pair which is tested         */
  double Qmin;                        /* current minimun of the criterion */
  
  Qmin=1.0e300;
  *a = *b = -1;
  for(i=0; i < n; i++)
    {
      if (56 == i)
	{
	  /*printf("Debug r is %d, i is %d Emptied[i] is %d\n",r,i,Emptied[i]);*/
	}
      if(0 == Emptied[i])
        {
	  if (-1 == *a)
	    *a = i;
	  else if (-1 == *b)
	    *b = i;
	  for(j=0; j < i; j++)
	    {
	      if(0 == Emptied[j])
		{
		  Qxy=Agglomerative_criterion(i,j,delta,r);
		  if(Qxy < Qmin)
		    {
		      Qmin=Qxy;
		      *a=i;          
		      *b=j;
		    }
		}  
	    }
        }
    }
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;Sum_S;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;  Description : This function retrieves the sum Sx from the diagonal       ;
;                of the delta matrix.                                       ;
;                                                                           ;
;  input       :                                                            ;
;               int i          : subtree i                                  ;
;               double **delta : the delta matrix                            ;
;                                                                           ;
;  return value:                                                            ;
;                double delta[i][i] : sum Si                                 ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

double Sum_S(int i, double **delta)          /* get sum Si form the diagonal */
{
  return(delta[i][i]);
}

/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;   
;                          Formulae                                         ;
;                                                                           ;  
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/


double Agglomerative_criterion(int i, int j, double **delta, int r)
{
  double Qij;
  Qij=(r-2)*delta[i][j]                           /* Formula (1) */
    -Sum_S(i,delta)
    -Sum_S(j,delta); 
  
  return(Qij);                       
}


double Branch_length(int a, int b, double **delta, int r)
{
  double length;
  length=0.5*(delta[a][b]                         /* Formula (2) */
	      +(Sum_S(a,delta)
		-Sum_S(b,delta))/(r-2)); 
  return(length);                                   
}


double Reduction4(int a, double la, int b, double lb, int i,
		 double **delta)
{
  double Dui;
  Dui=0.5*(delta[a][i]-la + delta[b][i]-lb);                /* Formula (4) */
  return(Dui);
}


/*;;;;;;;;;;;;;;;;;;;;;;Finish_branch_length;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;  Description :  Compute the length of the branch attached                 ;
;                 to the subtree i, during the final step                   ;
;                                                                           ;
;  input       :                                                            ;
;                int i          : position of subtree i                     ;
;                int j          : position of subtree j                     ;
;                int k          : position of subtree k                     ;
;                double **delta :                                            ;
;                                                                           ;
;  return value:                                                            ;
;                double length  : The length of the branch                   ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

double Finish_branch_length(int i, int j, int k, double **delta)
{
  double length;
  length=0.5*(delta[i][j] + delta[i][k] - delta[j][k]);
  return(length);
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;Finish;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;  Description : This function compute the length of the lasts three        ;
;                subtrees and write the tree in the output file.            ;
;                                                                           ;
;  input       :                                                            ;
;                double **delta  : the delta matrix                         ;
;                int N           : the total number of nodes in the tree    ;
;                int *Emptied    : array tracking which nodes are OTUs      ;
;                                                                           ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

tree *Finish(double **delta, int N, int *Emptied, node **nodeArray)
{
  tree *NJTree;
  node *root,*nleft,*ncenter,*nright;
  char nodeLabel[NODE_LABEL_LENGTH];
  char edgeLabel[EDGE_LABEL_LENGTH];
  int l=0;
  int i=0;
  double length;
  edge *left,*right,*center;
  int last[3];                            /* the last three subtrees     */
  
  NJTree = newTree();
  sprintf(nodeLabel,"",N-1);
  root=makeNewNode(nodeLabel,N-1);
  

  while(l < N)
    {                                       /* find the last tree subtree  */
      if(0==Emptied[l])
	{
	  /*printf("Debug: Finish, i is %d.\n",i);*/
	  last[i]=l;
	  i++;
	}
      l++;
    }
  
  nleft = nodeArray[last[0]];
  sprintf(edgeLabel,"E%d",N-3);
  length=Finish_branch_length(last[0],last[1],last[2],delta);  
  left=makeEdge(edgeLabel,root,nleft,length);
  root->leftEdge = nleft->parentEdge = left;

  nright = nodeArray[last[1]];
  sprintf(edgeLabel,"E%d",N-2);
  length=Finish_branch_length(last[1],last[0],last[2],delta);  
  right=makeEdge(edgeLabel,root,nright,length);
  root->rightEdge = nright->parentEdge = right;
  
  ncenter = nodeArray[last[2]];
  sprintf(edgeLabel,"E%d",N-1);
  length=Finish_branch_length(last[2],last[1],last[0],delta);  
  center=makeEdge(edgeLabel,root,ncenter,length);
  root->middleEdge=ncenter->parentEdge=center;
  
  NJTree->root = root;
  NJTree->size=N;
  return(NJTree);

}
